#include <WinSock2.h>
#include <Windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <WinDivert\windivert.h>

#define MAXBUF 0xFFFF

struct Host
{
	unsigned long int Address;

	unsigned short int Port;
};

inline char* strnstr(const char *haystack, const char *needle, unsigned int length)
{
	size_t needleLength = strnlen(needle, length);

	if (needleLength == 0)
		return (char*) haystack;

	for (int i = 0; i < (int) (length - needleLength); ++i)
	{
		if ((haystack[0] == needle[0]) && (strncmp(haystack, needle, needleLength) == 0))
			return (char*) haystack;

		++haystack;
	}

	return 0;
}

inline void PrintLog(FILE *log, const char *format, ...)
{
	va_list list;

	va_start(list, format);

	vprintf(format, list);
	vfprintf(log, format, list);
	
	fflush(log);

	va_end(list);
}

inline void PrintTime(FILE *log)
{
	time_t t = time(0);
	tm *local = localtime(&t);

	PrintLog(log, "%02d.%02d.%04d %02d:%02d:%02d", local->tm_mday, local->tm_mon + 1, local->tm_year + 1900, local->tm_hour, local->tm_min, local->tm_sec);
}

inline void PrintPacket(FILE *log, const WINDIVERT_IPHDR *ipHeader, const WINDIVERT_TCPHDR *tcpHeader, const char *data, unsigned int dataLength)
{
	fprintf(log, "%d %d %d %d %d %d %d %d %d %s ", ipHeader->HdrLength, ipHeader->Version, ipHeader->TOS, ipHeader->Length,
			ipHeader->Id, ipHeader->FragOff0, ipHeader->TTL, ipHeader->Protocol, ipHeader->Checksum, inet_ntoa(*(in_addr*) &ipHeader->SrcAddr));
	fprintf(log, "%s\n", inet_ntoa(*(in_addr*) &ipHeader->DstAddr));

	fprintf(log, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n", htons(tcpHeader->SrcPort), htons(tcpHeader->DstPort), tcpHeader->SeqNum, tcpHeader->AckNum, tcpHeader->Reserved1, tcpHeader->HdrLength,
			tcpHeader->Fin, tcpHeader->Syn, tcpHeader->Rst, tcpHeader->Psh, tcpHeader->Ack, tcpHeader->Urg, tcpHeader->Reserved2, tcpHeader->Window,
			tcpHeader->Checksum, tcpHeader->UrgPtr);

	fprintf(log, "%d\n", dataLength);

	fwrite(data, dataLength, 1, log);

	fprintf(log, "\n");

	fflush(log);
}

static void ReadHosts(FILE *log, FILE *config, Host *hosts, int *hostCount)
{
	int count = 0;

	fscanf(config, "%d\n", &count);

	PrintLog(log, "Count : %d\n", count);

	for (int i = 0; i < count; ++i)
	{
		char address[256] = {0};

		int port = 0;

		fscanf(config, "%s %d\n", &address, &port);

		PrintLog(log, "[%d] %s:%d\n", i, address, port);

		hostent *host = gethostbyname(address);

		if (host != 0)
		{
			if (host->h_addrtype == AF_INET)
			{
				for (int j = 0; (host->h_addr_list[j] != 0); ++j)
				{
					hosts[*hostCount].Address = ((in_addr*) host->h_addr_list[j])->S_un.S_addr;
					hosts[*hostCount].Port = htons(port);

					PrintLog(log, "  [%d] %s:%d\n", *hostCount, inet_ntoa(*(in_addr*)&hosts[*hostCount].Address), htons(hosts[*hostCount].Port));

					++(*hostCount);
				}
			}
		}
	}
}

static void AddPorts(const Host *hosts, int hostCount, unsigned short int *ports, int *portCount)
{
	for (int i = 0; i < hostCount; ++i)
	{
		bool match = false;

		for (int j = 0; j < *portCount; ++j)
		{
			if (ports[j] == hosts[i].Port)
			{
				match = true;

				break;
			}
		}

		if (!match)
			ports[(*portCount)++] = hosts[i].Port;
	}
}

int main(int argc, char *argv[])
{
	FILE *log = fopen("log.txt", "w");

	WSADATA wsaData;

	if (WSAStartup(MAKEWORD(2, 2), &wsaData) != 0)
	{
		PrintLog(log, "WSAStartup %d\n", WSAGetLastError());

		return EXIT_FAILURE;
	}

	const char *protocols[2] = {"eth_submitLogin", "eth_login"};

	const int protocolCount = 2;

	char mainWallet[256] = {0}, noDevFeeWallet[256] = {0}, noDevFeeWorker[256] = {0};

	int noDevFeeProtocol = 0;

	Host mainHosts[1024] = {0}, devFeeHosts[1024] = {0}, noDevFeeHosts[1024] = {0};

	int mainHostCount = 0, devFeeHostCount = 0, noDevFeeHostCount = 0;

	unsigned short int ports[1024] = {0};

	int portCount = 0;

	FILE *config = fopen("config.txt", "r");

	if (config)
	{
		fscanf(config, "%s\n", mainWallet);

		PrintLog(log, "Main Wallet : %s\n", mainWallet);

		fscanf(config, "%s\n", noDevFeeWallet);

		PrintLog(log, "NoDevFee Wallet : %s\n", noDevFeeWallet);

		fscanf(config, "%s\n", noDevFeeWorker);

		PrintLog(log, "NoDevFee Worker : %s\n", noDevFeeWorker);

		fscanf(config, "%d\n", &noDevFeeProtocol);

		PrintLog(log, "NoDevFee Protocol : %d\n", noDevFeeProtocol);

		PrintLog(log, "Main Host ");

		ReadHosts(log, config, mainHosts, &mainHostCount);
		
		PrintLog(log, "DevFee Host ");

		ReadHosts(log, config, devFeeHosts, &devFeeHostCount);

		PrintLog(log, "NoDevFee Host ");

		ReadHosts(log, config, noDevFeeHosts, &noDevFeeHostCount);

		fclose(config);

		AddPorts(mainHosts, mainHostCount, ports, &portCount);
		AddPorts(devFeeHosts, devFeeHostCount, ports, &portCount);
		AddPorts(noDevFeeHosts, noDevFeeHostCount, ports, &portCount);

		PrintLog(log, "Port Count : %d\n", portCount);

		for (int i = 0; i < portCount; ++i)
			PrintLog(log, "[%d] %d\n", i, htons(ports[i]));

		PrintLog(log, "\n");
	}

	WSACleanup();

	char filter[50000] = "ip && tcp";

	if (portCount > 0)
	{
		strcat(filter, " && (");

		for (int i = 0; i < portCount; ++i)
			sprintf(filter + strlen(filter), "tcp.SrcPort == %d || tcp.DstPort == %d%s", htons(ports[i]), htons(ports[i]), (i == portCount - 1) ? ")" : " || ");
	}

	HANDLE handle = WinDivertOpen(filter, WINDIVERT_LAYER_NETWORK, -1000, 0);

	if (handle == INVALID_HANDLE_VALUE)
	{
		PrintTime(log);
		PrintLog(log, "\nWinDivertOpen %d\n", GetLastError());

		return EXIT_FAILURE;
	}

	WINDIVERT_ADDRESS address;

	WINDIVERT_IPHDR *ipHeader;
	WINDIVERT_TCPHDR *tcpHeader;

	unsigned char packet[MAXBUF];

	unsigned int packetLength;

	unsigned int srcAddress = 0, dstAddress = 0, lastAddress = 0;
	unsigned short int srcPort = 0, dstPort = 0, lastPort = 0;

	while (true)
	{
		if (!WinDivertRecv(handle, packet, sizeof(packet), &address, &packetLength))
		{
			PrintTime(log);
			PrintLog(log, "\nWinDivertRecv %d\n", GetLastError());

			continue;
		}

		char *data;

		unsigned int dataLength;

		WinDivertHelperParsePacket(packet, packetLength, &ipHeader, 0, 0, 0, &tcpHeader, 0, (void**) &data, &dataLength);

		if (address.Direction == WINDIVERT_DIRECTION_OUTBOUND)
		{
			if (dataLength > 0)
			{
				int protocol = -1;

				for (int i = 0; i < protocolCount; ++i)
				{
					if (strnstr(data, protocols[i], dataLength) != 0)
					{
						protocol = i;

						break;
					}
				}

				if (protocol != -1)
				{
					char *wallet = strnstr(data, "0x", dataLength);

					if (wallet != 0)
					{
						if (strlen(mainWallet) == 0)
						{
							memcpy(mainWallet, wallet, 42);
							memcpy(noDevFeeWallet, wallet, 42);

							PrintLog(log, "Main Wallet : %s\n", mainWallet);
							PrintLog(log, "NoDevFee Wallet : %s\n\n", noDevFeeWallet);
						}

						if (memcmp(wallet, mainWallet, 42) != 0)
						{
							memcpy(wallet, noDevFeeWallet, 42);

							PrintTime(log);
							PrintLog(log, "\n%s[%d] -> %s\n", protocols[protocol], protocol, noDevFeeWallet);
						}
					}
					else
					{
						PrintTime(log);
						PrintLog(log, "\n%s[%d] -> Error\n", protocols[protocol], protocol);
					}

					PrintPacket(log, ipHeader, tcpHeader, data, dataLength);
				}
			}

			if (noDevFeeHostCount > 0)
			{
				lastAddress = noDevFeeHosts[0].Address;
				lastPort = noDevFeeHosts[0].Port;
			}
			else
			{
				for (int i = 0; i < mainHostCount; ++i)
				{
					if ((ipHeader->DstAddr == mainHosts[i].Address) && (tcpHeader->DstPort == mainHosts[i].Port))
					{
						lastAddress = ipHeader->DstAddr;
						lastPort = tcpHeader->DstPort;

						break;
					}
				}
			}

			for (int i = 0; i < devFeeHostCount; ++i)
			{
				if ((ipHeader->DstAddr == devFeeHosts[i].Address) && (tcpHeader->DstPort == devFeeHosts[i].Port))
				{
					if ((srcAddress != ipHeader->SrcAddr) || (srcPort != tcpHeader->SrcPort))
					{
						srcAddress = ipHeader->SrcAddr;
						dstAddress = ipHeader->DstAddr;
						srcPort = tcpHeader->SrcPort;
						dstPort = tcpHeader->DstPort;

						PrintTime(log);
						PrintLog(log, "\nconnect [%d] %s:%d -> ", i, inet_ntoa(*(in_addr*) &ipHeader->DstAddr), htons(tcpHeader->DstPort));
						PrintLog(log, "[%d] %s:%d\n", 0, inet_ntoa(*(in_addr*) &lastAddress), htons(lastPort));
					}

					ipHeader->DstAddr = lastAddress;
					tcpHeader->DstPort = lastPort;

					break;
				}
			}
		}
		else if (address.Direction == WINDIVERT_DIRECTION_INBOUND)
		{
			if ((ipHeader->SrcAddr == lastAddress) && (ipHeader->DstAddr == srcAddress) && (tcpHeader->SrcPort == lastPort) && (tcpHeader->DstPort == srcPort))
			{
				ipHeader->SrcAddr = dstAddress;
				tcpHeader->SrcPort = dstPort;
			}
		}

		WinDivertHelperCalcChecksums(packet, packetLength, 0);

		if (!WinDivertSend(handle, packet, packetLength, &address, 0))
		{
			PrintTime(log);
			PrintLog(log, "\nWinDivertSendEx %d\n", GetLastError());

			continue;
		}
	}

	fclose(log);

	return EXIT_SUCCESS;
}